--[[
	/#########################################\
	##                                       ##
	##   Schattenbahnhofssteuerung fr EEP   ##
	##                                       ##
	##   Autor: Benjamin Hogl (BH2),         ##
	##          benjamin.hogl@gmx.de         ##
	##   Version 0.1, 2014-08-03             ##
	##                                       ##
	\#########################################/

Eine Schattenbahnhofssteuerung mit LUA fr EEP-Anlagen
    viele Mglichkeiten, aber einfach zu verwenden
]]

EEPGetElem = EEPGetElem or function(id)
	local signalRes=EEPGetSignal(id)
	if signalRes>0 then return signalRes
	else return EEPGetSwitch(id) end
end
EEPSetElem = EEPSetElem or function(id,pos,useCallback)
	if useCallback then
		local signalRes=EEPSetSignal(id,pos,useCallback)
		if signalRes>0 then return signalRes
		else return EEPSetSwitch(id,pos,useCallback) end
	else
		local signalRes=EEPSetSignal(id,pos)
		if signalRes>0 then return signalRes
		else return EEPSetSwitch(id,pos) end
	end
end

local Schattenbahnhof = {
	sbf={
		tracks={},
		debugLevel=1, -- 0: no output, 1: only errors, 2: errors and infos
		name="",
	},
	version={0,1},
}

function Schattenbahnhof:new(o,verbose)
	local res = o or {}
	setmetatable(res, self.sbf)
	self.sbf.__index = self.sbf
	if res.name ~= "" then res.name = " "..res.name end
	if res.create then
		if #res.tracks<1 and not (res.create.numTracks and res.create.signalIdStart and res.create.switchIdStart) then
			print('##################################################')
			print('## Fehler beim Anlagen von Schattenbahnhof',res.name,":")
			print('## Die Felder create.numTracks, create.signalIdStart')
			print('## und create.switchIdStart mssen vorhanden sein')
			print('##################################################')
			return {}
		end
		res.create.numTracks = res.create.numTracks or #res.tracks
		res.create.signalIdIncrement = res.create.signalIdIncrement or 1
		res.create.blockIdIncrement = res.create.blockIdIncrement or 1
		res.create.readyIdIncrement = res.create.readyIdIncrement or 1
		res.create.switchIdIncrement = res.create.switchIdIncrement or 1
		res.create.switchPosToTrack = res.create.switchPosToTrack or 2 -- "Abzweig"
		res.tracks=res.tracks or {}
		for trackNo=1,res.create.numTracks do
			res.tracks[trackNo]=res.tracks[trackNo] or {}
			if res.create.signalIdStart then res.tracks[trackNo].signalID = res.create.signalIdStart + (trackNo-1)*res.create.signalIdIncrement end
			if res.create.blockIdStart then res.tracks[trackNo].blockID = res.create.blockIdStart + (trackNo-1)*res.create.blockIdIncrement end
			if res.create.readyIdStart then res.tracks[trackNo].readyID = res.create.readyIdStart + (trackNo-1)*res.create.readyIdIncrement end
			local switches = {}
			for switchNo=1,trackNo do
				switches[switchNo]=
					{
						id=res.create.switchIdStart + (switchNo-1)*res.create.switchIdIncrement,
						pos=(res.create.switchPosToTrack%2)+1
					}
			end
			switches[trackNo].pos=res.create.switchPosToTrack -- letzte Weiche auf Abzweig
			if res.create.flipFirstSwitch then switches[1].pos=(switches[1].pos%2)+1 end -- erste Weiche bei Bedarf nochmal umdrehen
			if trackNo == res.create.numTracks and not res.create.switchForLastTrack then switches[trackNo]=nil end -- das letzte Gleis hat keine zustzliche Weiche
			if res.tracks[trackNo].switches == nil then
				res.tracks[trackNo].switches = switches
			else
				-- append entries to switch table
				for _,entry in pairs(switches) do
					res.tracks[trackNo].switches[#res.tracks[trackNo].switches+1]=entry
				end
			end
		end
		res.create=nil
	end
	
	-- check for correct syntax, correct it otherwise
	for trackNo,track in ipairs(res.tracks) do
		local newEntries={}
		for switchindex,switch in pairs(track.switches) do
			if type(switch)=="table" then
				if switch[1] then -- assume this format: switch={{423,1},{424,2},{id,stellung}}
					switch.id=switch[1]
					switch.pos=switch[2]
					switch[1]=nil
					switch[2]=nil
				end
			else -- assume this format: switch={[423]=1,[424]=2,[id]=stellung} 
				newEntries[#newEntries+1]={id=switchindex,pos=switch}
				track.switches[switchindex]=nil -- delete old entry
			end
		end
		-- append new entries to switch table
		for _,entry in pairs(newEntries) do
			track.switches[#track.switches+1]=entry
		end
		track.name = track.name or "Gleis "..trackNo	-- if no track name was given, assume "Gleis #"
		track.signalPosGo	= track.signalPosGo or 1
		track.signalPosStop	= track.signalPosStop or 2
		if not track.blockID then
			track.blockID = track.signalID
			track.blockPosFree = track.signalPosGo					
			track.blockPosBlocked = 0						-- 0 is no valid signal position, so no change will happen
		end
		if not track.readyID then
			track.readyID = track.signalID
			track.readyPosReady = track.signalPosStop
			track.readyPosWait = 0						-- 0 is no valid signal position, so no change will happen
		end
		-- if still not defined:
		track.blockPosBlocked = track.blockPosBlocked or 2
		track.blockPosFree = track.blockPosFree or 1
		track.readyPosReady = track.readyPosReady or 1
		track.readyPosWait = track.readyPosWait or 2
	end
	return res
end

function Schattenbahnhof.sbf:print(includeSwitches) 
	print()
	print("Schattenbahnhof",self.name," mit ",#self.tracks," Schattenbahnhofs-Gleisen: (Debug-Level ",self.debugLevel,")")
	for trackNo,track in ipairs(self.tracks) do
		local weichen=includeSwitches and ": " or ""
		if includeSwitches then
			for _,switch in ipairs(track.switches) do
				weichen=weichen.."ID:"..switch.id.."->"..switch.pos.." "
			end
		end
		print(track.name, "  |  ",(EEPGetSignal(track.blockID)==track.blockPosFree and "___   frei   " or "## belegt"),
			"  |  Signal-ID: ", track.signalID, ", Block-ID: ", track.blockID, ", Ready-ID: ", track.readyID, "  |  ", #track.switches, " Weichen",weichen)
	end
	print()
end

function Schattenbahnhof.sbf:einfahrt()
	for _,track in ipairs(self.tracks) do							-- Die Gleise der Reihe nach durchgehen
		if EEPGetElem(track.blockID)==track.blockPosFree then		-- Wenn das Gleis als frei markiert ist,
			EEPSetElem(track.blockID,track.blockPosBlocked,1)		-- Gleis als besetzt markieren und
			EEPSetElem(track.signalID,track.signalPosStop,1)		-- Signal auf Halt stellen
			for _,switch in ipairs(track.switches) do
				EEPSetElem(switch.id,switch.pos,1)					-- Weichen stellen
			end
			if self.debugLevel>=2 then print("[Schattenbahnhof",self.name,"] Einfahrt nach ",track.name) end
			return true
		end
	end
	-- Schleife ohne Ergebnis durchgelaufen
	if self.debugLevel>=1 then print("[Schattenbahnhof",self.name,"] Fehler: Schattenbahnhof voll, keine Einfahrt mglich") end
	return false
end

function Schattenbahnhof.sbf:ausfahrt()
	local readyTracks={}
	local num=0
	for trackNo,track in ipairs(self.tracks) do
		if EEPGetElem(track.readyID)==track.readyPosReady then		-- Gleis bereit zur Abfahrt?
			num=num+1
			readyTracks[num]=trackNo								-- zur Liste der mglichen Gleise hinzufgen
		end
	end
	if num<1 then
		if self.debugLevel>=1 then print("[Schattenbahnhof",self.name,"] Fehler: Schattenbahnhof ist leer, keine Ausfahrt mglich") end
		return false
	end
	local rand=math.random(num)
	local selectedTrack = self.tracks[readyTracks[rand]]			-- Zufllig eines der belegten Gleise auswhlen
	EEPSetElem(selectedTrack.signalID,selectedTrack.signalPosGo,1)	-- Ausfahrt freigeben und Gleis als frei markieren
	EEPSetElem(selectedTrack.readyID,selectedTrack.readyPosWait,1)	-- Keine weitere Ausfahrt mglich
	if self.debugLevel>=2 then print("[Schattenbahnhof",self.name,"] Ausfahrt aus ",selectedTrack.name) end
	return true
end


print("Schattenbahnhofssteuerung_BH2 v",table.concat(Schattenbahnhof.version,'.')," wurde eingebunden")

return Schattenbahnhof